/*-----------------------------------------------------------------------------

 2002, Jeff M Soft und Hardware GmbH, All Rights Reserved

-------------------------------------------------------------------------------

	    v - bow position

---b0------>*----b1----->  out
<--b2-------*<---b3------
			|^
		     b4	|| b5 (bridge vertical)
			v|
	v-<----b6-----<-*->---------b7---->-v
	->-(one path)-->*<------(one path)-<-

<----l1---->|<----l2--->^
			l3
			v
	|<------l4----->|<-----l5------->|

there are no l4 and l5 variables used, only i (int length) to reduce computations

values of l4 and l5 are of course twice the graphic because these are employed as recursive paths.
these paths are "air columns" in the acoustic body. the ratio between the lengths is applied
to transmission coefficient. these paths also have two read pointers each (2nd at 70% of total len)


(small diagram for copy/paste when coding paths)
b0>b1>
b2<b3<
   b4 v^ b5
   b6<><>b7
l1 l2
      l3
   l4    l5

bowing junction implicated from julius o. smith III's info on 'friedlander-keller diagram'

(input velocity - string velocity) is transformed by this friction modeling curve:

                       |     .
                       |    /|\
                       |   / | '.
                       |  /  |   '-.
                       | /   |      ''-..
                       |/    |           '''---
(-)--------------------0-----+--------------(+)
---...                /|     |
      ''-..          / |     |
           '-.      /  |     |
              '.   /   |     |
                \ /    |     |
                 '     |     * release point

in the center band output is linear.. the string and bow are sticking together, the bow is 100% pushing the string.
slippage increases with velocity difference, reducing the bow's effect on the string

note the general shape

my v0 = jos3 v^, abs(v0) = v+^
my v1 = ~jos3 r
*/

#include "math.h"
#include "Module.h"
#include "SEMod_struct.h"
#include "SEPin.h"
#include <assert.h>

// define some constants to make referencing in/outs clearer
#define PN_INPUT1 0
#define PN_INPUT2 1
#define PN_INPUT3 2
#define PN_INPUT4 3
#define PN_INPUT5 4
#define PN_INPUT6 5
#define PN_INPUT7 6
#define PN_INPUT8 7
#define PN_INPUT9 8
#define PN_INPUT10 9
#define PN_INPUT11 10
#define PN_INPUT12 11
#define PN_INPUT13 12
#define PN_INPUT14 13
#define PN_INPUT15 14
#define PN_INPUT16 15

#define PN_OUTPUT1 16
#define PN_OUTPUT2 17
static const float anti_denormal = 1e-18f;
inline void kill_denormal(float &p_sample ) { p_sample += anti_denormal;p_sample -= anti_denormal;};

void Module::sub_process(long buffer_offset, long sampleFrames )
{
    float *in1  = buffer_offset + input1_buffer;
    float *in2  = buffer_offset + input2_buffer;
    float *in3  = buffer_offset + input3_buffer;
    float *in4  = buffer_offset + input4_buffer;
    float *in5  = buffer_offset + input5_buffer;
    float *in6  = buffer_offset + input6_buffer;
    float *in7  = buffer_offset + input7_buffer;
    float *in8  = buffer_offset + input8_buffer;
    float *in9  = buffer_offset + input9_buffer;
    float *in10  = buffer_offset + input10_buffer;
    float *in11  = buffer_offset + input11_buffer;
    float *in12  = buffer_offset + input12_buffer;
    float *in13  = buffer_offset + input13_buffer;
    float *in14  = buffer_offset + input14_buffer;
    float *in15  = buffer_offset + input15_buffer;
    float *in16  = buffer_offset + input16_buffer;

    float *out1 = buffer_offset + output1_buffer;
    float *out2 = buffer_offset + output2_buffer;
    while(--sampleFrames >= 0) {

	if (*in5 != c1) {
		c1 = *in5;
		c2 = *in4;
		l0 = .5f * (samplerate / (440.f * pow(2.f, (*in5 * 10.f) - 5.f)));	// calc total length
		if (l0 < 2.f) l0 = 2.f;
		else if (l0 > 8191.f) l0 = 8191.f;

		l0 *= (1.f - *in8 * .04f);

		l1 = l0;
		l2 = *in4 * 50.f; // l1 * *in4;
		if (l2 > l1 - 1.f) l2 = l1 - 1.f;
		else if (l2 < 1.f) l2 = 1.f;
		l1 -= l2;
		i1 = l1;
		i2 = l2;
		d1 = l1 - i1;
		d2 = l2 - i2;
	}
	else if (*in4 != c2) {
		c2 = *in4;
		l1 = l0;
		l2 = *in4 * 50.f; // l1 * *in4;
		if (l2 > l1 - 1.f) l2 = l1 - 1.f;
		else if (l2 < 1.f) l2 = 1.f;
		l1 -= l2;
		i1 = l1;
		i2 = l2;
		d1 = l1 - i1;
		d2 = l2 - i2;
	}
	if ((*in13 != c3) || (*in14 != c4) || (*in15 != c5)) {
		c3 = *in13;
		c4 = *in14;
		c5 = *in15;
		l3 = *in13 * 100.f;
		if (l3 < 1.f) l3 = 1.f;
		else if (l3 > 8191.f) l3 = 8191.f;
		i3 = l3;
		d3 = i3 - l3;

		rr = *in15 + 1.f; // *in15 input 0.f - 1.f
		if (rr < 1.f) rr = 1.f;
		else if (rr > 2.f) rr = 2.f;
		rr *= .70710678f;
		rr *= rr;	// outputs appx. .5 to 2.0 with center close to 1
		rt = *in14 * (samplerate / 190.f);
		i4 = rt * rr;
		i5 = rt;
		if (i4 < 1) i4 = 1;
		else if (i4 > 8191) i4 = 8191;
		if (i5 < 1) i5 = 1;
		else if (i5 > 8191) i5 = 8191;
		rr = ((rt * rr) - rt) / ((rt * rr) + rt);
		i6 = float((rt * rr) * .7f);
		i7 = float(rt * .8f);
		if (i6 < 1) i6 = 1;
		else if (i6 > 8191) i6 = 8191;
		if (i7 < 1) i7 = 1;
		else if (i7 > 8191) i7 = 8191;
	}
	r1 = p1 - i1;
	r2 = p2 - i2;
	r3 = p3 - i3;
	r4 = p4 - i4;
	r5 = p5 - i5;
	r6 = p4 - i6;
	r7 = p5 - i7;
	q1 = r1 - 1;
	q2 = r2 - 1;
	q3 = r3 - 1;
	if (r1 < 0) r1 += 8192;
	if (r2 < 0) r2 += 8192;
	if (r3 < 0) r3 += 8192;
	if (r4 < 0) r4 += 8192;
	if (r5 < 0) r5 += 8192;
	if (r6 < 0) r6 += 8192;
	if (r7 < 0) r7 += 8192;
	if (q1 < 0) q1 += 8192;
	if (q2 < 0) q2 += 8192;
	if (q3 < 0) q3 += 8192;
	o0 = b0[r1] + (b0[q1] - b0[r1]) * d1;
	o1 = b1[r2] + (b1[q2] - b1[r2]) * d2;
	o2 = b2[r1] + (b2[q1] - b2[r1]) * d1;
	o3 = b3[r2] + (b3[q2] - b3[r2]) * d2;
	o4 = b4[r3] + (b4[q3] - b4[r3]) * d3;
	o5 = b5[r3] + (b5[q3] - b5[r3]) * d3;
	o6 = (b6[r4] + b6[r6]) * -.5f;
	o7 = (b7[r5] + b7[r7]) * -.5f;

	if (o4 < -1.f) o4 = -1.f;
	else if (o4 > 1.f) o4 = 1.f;

	//o0 -= *in8 * (o0 - f0);
	o3 -= *in8 * (o3 - f3);
	o5 -= *in10 * (o5 - f5);
	o7 -= *in12 * (o7 - f7);
	//f0 = o0;
	f3 = o3;
	f5 = o5;
	f7 = o7;

	kill_denormal(o0);
	kill_denormal(o5);
	kill_denormal(o7);

	o1 *= *in7;				// try commenting this out
	o2 *= *in7;
	o4 *= *in9;
	o6 *= *in11;

	v0 = *in2 - (o0 + o3);
	v1 = (v0 * *in3) / *in16;
	if (abs(v0) < abs(v1)) v1 = v0;
	v1 *= abs(v0);
	v2 = v1 / (v1 + 15.f);
	v2 *= abs(v0) * *in1;

	//if (v2 < -1.f) v2 = -1.f;
	//else if (v2 > 1.f) v2 = 1.f;

	b1[p2] = o0 + v2 + *in6;
	b2[p1] = o3 + v2 + *in6;

	b0[p1] = -o2;
	b3[p2] = o5 * -.5f - o1;
	b4[p3] = o1 * 1.f - o5;
	b5[p3] = (o6 + o7) * .7f - o4;

	r = rr * (o6 - o7);
	b6[p4] = o4 * .7f + o7 + r;
	b7[p5] = o4 * .7f + o6 + r;

	*out1 = o1;
	*out2 = o7;

	p1++;
	p2++;
	p3++;
	p4++;
	p5++;
	p1 %= 8192;
	p2 %= 8192;
	p3 %= 8192;
	p4 %= 8192;
	p5 %= 8192;

	in1++;
	in2++;
	in3++;
	in4++;
	in5++;
	in6++;
	in7++;
	in8++;
	in9++;
	in10++;
	in11++;
	in12++;
	in13++;
	in14++;
	in15++;
	in16++;

	out1++;
	out2++;
    }
}

void Module::sub_process_static(long buffer_offset, long sampleFrames )
{
	sub_process(buffer_offset, sampleFrames);
	static_count = static_count - sampleFrames;
	if( static_count <= 0 )
	{
		CallHost(seaudioMasterSleepMode);
	}
}

Module::Module(seaudioMasterCallback2 seaudioMaster, void *p_resvd1) : SEModule_base(seaudioMaster, p_resvd1)
{
}

void Module::open()
{
	SEModule_base::open();
	SET_PROCESS_FUNC(Module::sub_process);
	zchannels();

	samplerate	= getSampleRate();

	l0 = l1 = l2 = l3 = 0.f;
	d1 = d2 = d3 = 0.f;
	c1 = c2 = c3 = c4 = c5 = 0.f;

	i1 = i2 = i3 = i4 = i5 = i6 = i7 = 0;
	p1 = p2 = p3 = p4 = p5 = 8191;
	q1 = q2 = q3 = 0;
	r1 = r2 = r3 = r4 = r5 = r6 = r7 = 0;
	r = rr = rt = 0.f;

	o0 = o1 = o2 = o3 = o4 = o5 = o6 = o7 = 0.f;
	f0 = f3 = f5 = f7 = 0.f;
	v0 = v1 = v2 = 0.f;

	getPin(PN_OUTPUT1)->TransmitStatusChange( SampleClock(), ST_RUN);
	getPin(PN_OUTPUT2)->TransmitStatusChange( SampleClock(), ST_RUN);
}

Module::~Module()
{
	//delete [] b;
	zchannels();
}

// describe your module
bool Module::getModuleProperties (SEModuleProperties* properties)
{
	properties->name = "boing";
	properties->id = "rl boing";
	properties->about = "boing by rurik leffanta\n5/26/07\n\nbowed string model\n\nb = bow\ns = string\nbr = bridge\nr = resonant space\n\n\n";
	return true;
}

// describe the pins (plugs and parameters)
bool Module::getPinProperties (long index, SEPinProperties* properties)
{
	switch( index )
	{
		// typical input plug (inputs are listed first)
	case 0:
		properties->name				= "b.force";
		properties->variable_address	= &input1_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 1:
		properties->name				= "b.vel";
		properties->variable_address	= &input2_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "10";
		break;
	case 2:
		properties->name				= "b.grip";
		properties->variable_address	= &input3_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "6";
		break;
	case 3:
		properties->name				= "b.pos";
		properties->variable_address	= &input4_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0.1";
		break;
	case 4:
		properties->name				= "pitch";
		properties->variable_address	= &input5_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "3";
		break;
	case 5:
		properties->name				= "in";
		properties->variable_address	= &input6_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 6:
		properties->name				= "s.feed";
		properties->variable_address	= &input7_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 7:
		properties->name				= "s.lpf";
		properties->variable_address	= &input8_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 8:
		properties->name				= "br.feed";
		properties->variable_address	= &input9_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 9:
		properties->name				= "br.lpf";
		properties->variable_address	= &input10_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 10:
		properties->name				= "r.feed";
		properties->variable_address	= &input11_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 11:
		properties->name				= "r.lpf";
		properties->variable_address	= &input12_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case 12:
		properties->name				= "br.len";
		properties->variable_address	= &input13_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "2";
		break;
	case 13:
		properties->name				= "r.len";
		properties->variable_address	= &input14_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "2";
		break;
	case 14:
		properties->name				= "r.ratio";
		properties->variable_address	= &input15_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "5";
		break;
	case 15:
		properties->name				= "s.imped";
		properties->variable_address	= &input16_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "6";
		break;

		// typical output plug
	case 16:
		properties->name				= "out";
		properties->variable_address	= &output1_buffer;
		properties->direction			= DR_OUT;
		properties->datatype			= DT_FSAMPLE;
		break;
	case 17:
		properties->name				= "r.out";
		properties->variable_address	= &output2_buffer;
		properties->direction			= DR_OUT;
		properties->datatype			= DT_FSAMPLE;
		break;
	default:
		return false; // host will ask for plugs 0,1,2,3 etc. return false to signal when done
	};
	return true;
}

// An input plug has changed value
void Module::OnPlugStateChange(SEPin *pin)
{
}

void Module::zchannels() {
	int i;
	for (i = 0; i < 8192; i++) {
		b0[i] = 0.f;
		b1[i] = 0.f;
		b2[i] = 0.f;
		b3[i] = 0.f;
		b4[i] = 0.f;
		b5[i] = 0.f;
		b6[i] = 0.f;
		b7[i] = 0.f;
	}
}